"""Collection of shared djLint functions."""
import regex as re

from .settings import Config


def is_ignored_block_opening(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = list(
        re.finditer(
            config.ignored_blocks_inline,
            item,
            flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE | re.DOTALL,
        )
    )

    if inline:
        last_index = inline[
            -1
        ].end()  # get the last index. The ignored opening should start after this.

    return bool(
        re.search(
            re.compile(config.ignored_block_opening, re.IGNORECASE | re.VERBOSE),
            item[last_index:],
        )
    )


def is_script_style_block_opening(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = list(
        re.finditer(
            config.script_style_inline,
            item,
            flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE | re.DOTALL,
        )
    )

    if inline:
        last_index = inline[
            -1
        ].end()  # get the last index. The ignored opening should start after this.

    return bool(
        re.search(
            re.compile(config.script_style_opening, re.IGNORECASE | re.VERBOSE),
            item[last_index:],
        )
    )


def inside_protected_trans_block(config: Config, html: str, match: re.Match) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.

    True = non indentable > inside ignored trans block
    False = indentable > either inside a trans trimmed block, or somewhere else, but not a trans non trimmed :)
    """
    last_index = 0
    close_block = re.search(
        re.compile(
            config.ignored_trans_blocks_closing, flags=re.IGNORECASE | re.VERBOSE
        ),
        match.group(),
    )

    if not close_block:
        return False

    non_trimmed = list(
        re.finditer(
            re.compile(
                config.ignored_trans_blocks,
                flags=re.IGNORECASE | re.VERBOSE | re.DOTALL,
            ),
            html,
        )
    )

    trimmed = list(
        re.finditer(
            re.compile(
                config.trans_trimmed_blocks,
                flags=re.IGNORECASE | re.VERBOSE | re.DOTALL,
            ),
            html,
        )
    )

    # who is max?
    if non_trimmed and (not trimmed or non_trimmed[-1].end() > trimmed[-1].end()):
        # non trimmed!
        # check that this is not an inline block.
        non_trimmed_inline = list(
            re.finditer(
                re.compile(
                    config.ignored_trans_blocks,
                    flags=re.IGNORECASE | re.VERBOSE | re.DOTALL,
                ),
                match.group(),
            )
        )

        if non_trimmed_inline:
            last_index = non_trimmed[
                -1
            ].end()  # get the last index. The ignored opening should start after this.

            return re.search(
                re.compile(
                    config.ignored_trans_blocks_closing,
                    flags=re.IGNORECASE | re.VERBOSE,
                ),
                html[last_index:],
            )

        return close_block.end(0) <= non_trimmed[-1].end()

    elif trimmed:
        # inside a trimmed block, we can return true to continue as if
        # this is a indentable block
        return close_block.end(0) > trimmed[-1].end()
    return False

    # print(close_block)
    # if non_trimmed:
    #     last_index = non_trimmed[
    #         -1
    #     ].end()  # get the last index. The ignored opening should start after this.

    # return re.search(
    #     re.compile(
    #         config.ignored_trans_blocks_closing, flags=re.IGNORECASE | re.VERBOSE
    #     ),
    #     html[last_index:],
    # )


def is_ignored_block_closing(config: Config, item: str) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = list(
        re.finditer(
            re.compile(config.ignored_inline_blocks, flags=re.IGNORECASE | re.VERBOSE),
            item,
        )
    )

    if inline:
        last_index = inline[
            -1
        ].end()  # get the last index. The ignored opening should start after this.

    return re.search(
        re.compile(config.ignored_block_closing, flags=re.IGNORECASE | re.VERBOSE),
        item[last_index:],
    )


def is_script_style_block_closing(config: Config, item: str) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = list(
        re.finditer(
            re.compile(config.script_style_inline, flags=re.IGNORECASE | re.VERBOSE),
            item,
        )
    )

    if inline:
        last_index = inline[
            -1
        ].end()  # get the last index. The ignored opening should start after this.

    return re.search(
        re.compile(config.script_style_closing, flags=re.IGNORECASE | re.VERBOSE),
        item[last_index:],
    )


def is_safe_closing_tag(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    last_index = 0
    inline = list(
        re.finditer(
            re.compile(
                config.ignored_inline_blocks + r" | " + config.ignored_blocks,
                flags=re.IGNORECASE | re.VERBOSE | re.MULTILINE | re.DOTALL,
            ),
            item,
        )
    )

    if inline:
        last_index = inline[
            -1
        ].end()  # get the last index. The ignored opening should start after this.

    return re.search(
        re.compile(config.safe_closing_tag, flags=re.IGNORECASE | re.VERBOSE),
        item[last_index:],
    )


def inside_template_block(config: Config, html: str, match: re.Match) -> bool:
    """Check if a re.Match is inside of a template block."""
    return any(
        ignored_match.start(0) <= match.start() and match.end(0) <= ignored_match.end()
        for ignored_match in list(
            re.finditer(
                re.compile(
                    config.template_blocks,
                    re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
                ),
                html,
            )
        )
    )


def inside_ignored_linter_block(config: Config, html: str, match: re.Match) -> bool:
    """Check if a re.Match is inside of a ignored linter block."""
    return any(
        ignored_match.start(0) <= match.start() and match.end(0) <= ignored_match.end()
        for ignored_match in list(
            re.finditer(
                re.compile(
                    config.ignored_linter_blocks,
                    re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
                ),
                html,
            )
        )
    )


def inside_ignored_block(config: Config, html: str, match: re.Match) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        ignored_match.start(0) <= match.start() and match.end(0) <= ignored_match.end()
        for ignored_match in list(
            re.finditer(
                re.compile(
                    config.ignored_blocks,
                    re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
                ),
                html,
            )
        )
        + list(
            re.finditer(
                re.compile(config.ignored_inline_blocks, re.IGNORECASE | re.VERBOSE),
                html,
            )
        )
    )


def child_of_unformatted_block(config: Config, html: str, match: re.Match) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        ignored_match.start(0) < match.start() and match.end(0) <= ignored_match.end()
        for ignored_match in list(
            re.finditer(
                re.compile(
                    config.unformatted_blocks,
                    re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
                ),
                html,
            )
        )
    )


def child_of_ignored_block(config: Config, html: str, match: re.Match) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        ignored_match.start(0) < match.start() and match.end(0) <= ignored_match.end()
        for ignored_match in list(
            re.finditer(
                re.compile(
                    config.ignored_blocks,
                    re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE,
                ),
                html,
            )
        )
        + list(
            re.finditer(
                re.compile(config.ignored_inline_blocks, re.IGNORECASE | re.VERBOSE),
                html,
            )
        )
    )


def overlaps_ignored_block(config: Config, html: str, match: re.Match) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    return any(
        # don't require the match to be fully inside the ignored block.
        # poorly build html will probably span ignored blocks and should be ignored.
        (
            ignored_match.start(0) <= match.start()
            and match.start() <= ignored_match.end()
        )
        or (
            ignored_match.start(0) <= match.end() and match.end() <= ignored_match.end()
        )
        for ignored_match in list(
            re.finditer(
                re.compile(
                    config.ignored_blocks,
                    re.DOTALL | re.IGNORECASE | re.VERBOSE | re.MULTILINE | re.DOTALL,
                ),
                html,
            )
        )
        + list(
            re.finditer(
                re.compile(config.ignored_inline_blocks, re.IGNORECASE | re.VERBOSE),
                html,
            )
        )
    )


def inside_ignored_rule(config: Config, html: str, match: re.Match, rule: str) -> bool:
    """Check if match is inside an ignored pattern."""
    for rule_regex in config.ignored_rules:
        for ignored_match in list(
            re.finditer(
                re.compile(rule_regex, re.DOTALL | re.IGNORECASE | re.VERBOSE),
                html,
            )
        ):
            if (
                rule in list(set(re.split(r"\s|,", ignored_match.group(1).strip())))
                and (
                    ignored_match.start(0) <= match.start()
                    and match.start() <= ignored_match.end()
                )
                or not ignored_match.group(1).strip()
                and (
                    ignored_match.start(0) <= match.end()
                    and match.end() <= ignored_match.end()
                )
            ):
                return True
    return False



"""Condense HTML.

1. Put short html tags back on one line
2. Put short template tags back on one line
"""

from functools import partial

import regex as re

from ..helpers import (
    inside_ignored_block,
    inside_protected_trans_block,
    is_safe_closing_tag,
)
from ..settings import Config


def clean_whitespace(html: str, config: Config) -> str:
    """Compress back tags that do not need to be expanded."""
    # put empty tags on one line

    def strip_space(config: Config, html: str, match: re.Match) -> str:
        """Trim leading whitespace."""
        # either inside a block, or this is a newline + closing block tag.
        # if it is a newline + closing block we can format it.

        if inside_ignored_block(config, html, match) and not is_safe_closing_tag(
            config, match.group()
        ):
            return match.group()

        # trimmed blocks should not be here.
        # we need to full html to check what type of
        # opening block it was - trimmed or not trimmed
        if inside_protected_trans_block(config, html[: match.end()], match):
            return match.group().rstrip()

        lines = len(
            re.findall(
                r"\n",
                match.group(2),
            )
        )
        blank_lines = "\n" * lines
        if lines > config.max_blank_lines:
            blank_lines = "\n" * max(config.max_blank_lines, 0)
        return match.group(1) + blank_lines

    func = partial(strip_space, config, html)

    line_contents = r"(.*?)"
    trailing_contents = r"\n \t"

    if config.preserve_blank_lines:
        line_contents = r"([^\n]+?)"
        trailing_contents = r" \t"

    if not config.preserve_leading_space:
        # remove any leading/trailing space
        html = re.sub(
            re.compile(rf"^[ \t]*{line_contents}([{trailing_contents}]*)$", re.M),
            func,
            html,
        )

    else:
        # only remove leading space in front of tags
        # <, {%
        html = re.sub(
            re.compile(rf"^[ \t]*((?:<|{{%).*?)([{trailing_contents}]*)$", re.M),
            func,
            html,
        )
        html = re.sub(
            re.compile(rf"^{line_contents}([{trailing_contents}]*)$", re.M), func, html
        )

    def add_blank_line_after(config: Config, html: str, match: re.Match) -> str:
        """Add break after if not in ignored block."""
        if inside_ignored_block(config, html, match):
            return match.group()

        # check that next line is not blank.
        if html[match.end() : match.end() + 1] != "\n":  # noqa:E203
            return match.group() + "\n"

        return match.group()

    func = partial(add_blank_line_after, config, html)

    # should we add blank lines after load tags?
    if config.blank_line_after_tag:
        for tag in [x.strip() for x in config.blank_line_after_tag.split(",")]:
            html = re.sub(
                re.compile(
                    rf"((?:{{%\s*?{tag}\b[^}}]+?%}}\n?)+)",
                    re.IGNORECASE | re.MULTILINE | re.DOTALL,
                ),
                func,
                html,
            )

    def add_blank_line_before(config: Config, html: str, match: re.Match) -> str:
        """Add break before if not in ignored block and not first line in file."""
        if inside_ignored_block(config, html, match) or match.start() == 0:
            return match.group()

        return "\n" + match.group()

    func = partial(add_blank_line_before, config, html)

    # should we add blank lines before load tags?
    if config.blank_line_before_tag:
        for tag in [x.strip() for x in config.blank_line_before_tag.split(",")]:
            html = re.sub(
                re.compile(
                    rf"(?<!^\n)((?:{{%\s*?{tag}\b[^}}]+?%}}\n?)+)",
                    re.IGNORECASE | re.MULTILINE | re.DOTALL,
                ),
                func,
                html,
            )

    # add line after yaml front matter

    def yaml_add_blank_line_after(html: str, match: re.Match) -> str:
        """Add break after if not in ignored block."""
        if match.start() == 0 and not html.startswith("\n\n", match.end()):
            # verify there are not already blank lines
            return match.group() + "\n"

        return match.group()

    if config.no_line_after_yaml is False:
        func = partial(yaml_add_blank_line_after, html)
        html = re.sub(
            re.compile(
                r"(^---.+?---)$",
                re.MULTILINE | re.DOTALL,
            ),
            func,
            html,
        )

    return html


def condense_html(html, config):
    """Put short tags back on a single line."""
    if config.preserve_leading_space:
        # if a user is attempting to reuse any leading
        # space for other purposes, we should not try to remove it.
        return html

    def condense_line(config: Config, html: str, match: re.Match) -> str:
        """Put contents on a single line if below max line length."""
        if config.line_break_after_multiline_tag:
            # always force a break by pretending the line is too long.
            combined_length = config.max_line_length + 1
        else:
            combined_length = len(
                match.group(1).splitlines()[-1] + match.group(3) + match.group(4)
            )

        if (
            not inside_ignored_block(config, html, match)
            and combined_length < config.max_line_length
            and if_blank_line_after_match(config, match.group(3))
            and if_blank_line_before_match(config, match.group(3))
        ):
            return match.group(1) + match.group(3) + match.group(4)

        return match.group()

    def if_blank_line_after_match(config: Config, html: str) -> bool:
        """Check if there should be a blank line after."""
        if config.blank_line_after_tag:
            return not any(
                re.findall(
                    re.compile(
                        rf"((?:{{%\s*?{tag}[^}}]+?%}}\n?)+)",
                        re.IGNORECASE | re.MULTILINE | re.DOTALL,
                    ),
                    html,
                )
                for tag in [x.strip() for x in config.blank_line_after_tag.split(",")]
            )
        return True

    def if_blank_line_before_match(config: Config, html: str) -> bool:
        """Check if there should be a blank line before."""
        if config.blank_line_before_tag:
            return not any(
                re.findall(
                    re.compile(
                        rf"((?:{{%\s*?{tag}[^}}]+?%}}\n?)+)",
                        re.IGNORECASE | re.MULTILINE | re.DOTALL,
                    ),
                    html,
                )
                for tag in [x.strip() for x in config.blank_line_before_tag.split(",")]
            )
        return True

    # add blank lines before tags
    func = partial(condense_line, config, html)

    # put short single line tags on one line
    html = re.sub(
        re.compile(
            rf"(<({config.optional_single_line_html_tags})\b(?:\"[^\"]*\"|'[^']*'|{{[^}}]*}}|[^'\">{{}}])*>)\s*([^<\n]*?)\s*?(</(\2)>)",
            re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE,
        ),
        func,
        html,
    )

    # put short template tags back on one line. must have leading space
    # jinja +%} and {%+ intentionally omitted.
    html = re.sub(
        re.compile(
            rf"((?:\s|^){{%-?[ ]*?({config.optional_single_line_template_tags})\b(?:(?!\n|%}}).)*?%}})\s*([^%\n]*?)\s*?({{%-?[ ]+?end(\2)[ ]*?%}})",
            flags=re.IGNORECASE | re.MULTILINE | re.VERBOSE,
        ),
        func,
        html,
    )

    return html



"""rule H025: Check for orphans html tags."""
import copy
from typing import Any, Dict, List

import regex as re

from ..helpers import (
    inside_ignored_linter_block,
    inside_ignored_rule,
    overlaps_ignored_block,
)
from ..lint import get_line
from ..settings import Config


def run(
    rule: Dict[str, Any],
    config: Config,
    html: str,
    filepath: str,
    line_ends: List[Dict[str, int]],
    *args: Any,
    **kwargs: Any,
) -> List[Dict[str, str]]:
    """Check for orphans html tags."""
    errors: List[Dict[str, str]] = []
    open_tags: List[re.Match] = []
    orphan_tags: List[re.Match] = []

    for match in re.finditer(
        re.compile(
            r"<(/?(\w+))\s*(" + config.attribute_pattern + r"|\s*)*\s*?>",
            re.VERBOSE,
        ),
        html,
    ):
        if match.group(1) and not re.search(
            re.compile(rf"^/?{config.always_self_closing_html_tags}\b", re.I | re.X),
            match.group(1),
        ):
            # close tags should equal open tags
            if match.group(1)[0] != "/":
                open_tags.insert(0, match)
            else:
                for i, tag in enumerate(copy.deepcopy(open_tags)):
                    if tag.group(2) == match.group(1)[1:]:
                        open_tags.pop(i)
                        break
                else:
                    # there was no open tag matching the close tag
                    orphan_tags.append(match)

    for match in open_tags + orphan_tags:
        if (
            overlaps_ignored_block(config, html, match) is False
            and inside_ignored_rule(config, html, match, rule["name"]) is False
            and inside_ignored_linter_block(config, html, match) is False
        ):
            errors.append(
                {
                    "code": rule["name"],
                    "line": get_line(match.start(), line_ends),
                    "match": match.group().strip()[:20],
                    "message": rule["message"],
                }
            )
    return errors
